#!/bin/bash
#
source /etc/conf.d/vdr.admin.cfg
source /etc/conf.d/vdr


# These plugins will be used after installation
if [ "$(readlink /dev/cdrom)" != "" ] ; then
   DEFAULT_PLUGINS="mp3ng mplayer dvd cdda epgsearch burn skinenigmang admin "
else
   DEFAULT_PLUGINS="mp3ng mplayer epgsearch skinenigmang admin "
fi   

KV="$(uname -r)"
KV_MAJOR=`uname -r | cut -f1 -d "."`
KV_MINOR=`uname -r | cut -f2 -d "."`

MODCONF="/etc/modules.autoload.d/kernel-${KV_MAJOR}.${KV_MINOR}"

set_freevo() {
   sed -i /etc/freevo/freevo.conf -e "s/^geometry.*/geometry = $1/"
}

autoload() {
   LOADED=`cat $MODCONF | grep "^$*$"`
   if [ "$LOADED" = "" ] ; then
      echo "$*" >> $MODCONF
      echo "Adding $* to $MODCONF"
   else
      echo "$* is already in $MODCONF"
   fi
}

activate_plugin() {
   $SETVAL "PLG_$1" $2
}

test_directisa() {
   TEST_DATE="020200002002"
   hwclock --hctosys
   ACT_DATE=$(date +%m%d%H%M%Y.%S)
   date $TEST_DATE
   hwclock --systohc
   date $ACT_DATE
   hwclock --hctosys
   TD=$(date +%m%d%H%M%Y)
   date $ACT_DATE
   if [ "$TD" = "$TEST_DATE" ] ; then
      echo "RTC is working"
      DI=0
      hwclock --systohc
   else
      echo "directisa is necessary"   
      DI=1
   fi   
   $SETVAL "DIRECTISA_RTC" $DI
}

rm -f /_config/bin/rvdrinit.sh

# default volume = 150
sed -i /etc/vdr/setup.conf -e "s/^CurrentVolume = 255/CurrentVolume = 150/"
test_directisa
set_freevo "800x600"

cp /etc/vdr/plugins/admin/admin.conf.default /etc/vdr/plugins/admin/admin.conf
perl /etc/vdr/plugins/admin/setvdrconf.pl

perl /etc/vdr/plugins/admin/cfgplg.pl $DEFAULT_PLUGINS

sed -i /etc/freevo/local_conf.py -e "s/.*MPLAYER_VO_DEV.*/MPLAYER_VO_DEV = 'x11'/"
sed -i /etc/freevo/local_conf.py -e "s/.*XINE_VO_DEV.*/XINE_VO_DEV = 'xshm'/"
$SETVAL GUI Aus
$SETVAL LIRC ALDI4689-336
$SETVAL FRITZWATCH 0

sed -i $MODCONF -e "/^natsemi/d"
rc-update del activylircd >/dev/null 2>&1
$SETVAL WATCHDOG 60
$SETVAL SHUTDOWN_METHOD Halt
$SETVAL WAKEUP_METHOD ACPI
$SETVAL GUI_SWITCH_OUTPUT 0
$SETVAL GUI_LOOP_OUTPUT 0
$SETVAL ACPI 1

BOARD=$(/_config/bin/query_mb.sh)

case "$BOARD" in
   MP_*)
      $SETVAL MP_TVOUT 0
      $SETVAL WAKEUP_METHOD NVRAM
      $SETVAL SHUTDOWN_METHOD PowerOffKernel
      $SETVAL NVRAM_BOARD $BOARD
      $SETVAL ACPI 0
      echo "sh /_config/bin/mpkeys.sh" > /_config/bin/rvdrinit.sh
      chmod +x /_config/bin/rvdrinit.sh
      ;;
   EPIA_M)
      $SETVAL WAKEUP_METHOD NVRAM
      $SETVAL ACPI 1
      ;;
   ACTIVY)
      $SETVAL WAKEUP_METHOD STR
      $SETVAL ACPI 1
      $SETVAL LIRC ActivyFB
      $SETVAL GUI FreeVo
      autoload "i2c-dev"
      autoload "i2c-i801"
      autoload "natsemi"
      modprobe natsemi
      set_freevo "720x576"
      sed -i /etc/freevo/local_conf.py -e "s/.*MPLAYER_VO_DEV.*/MPLAYER_VO_DEV = 'x11'/"
      sed -i /etc/freevo/local_conf.py -e "s/.*XINE_VO_DEV.*/XINE_VO_DEV = 'x11'/"
      activate_plugin "alcd" "1"
      activate_plugin "dvd" "0"
      ;;
   *)
      ;;
esac

echo "Checking for ALSA ..."
sh /_config/bin/alsaconf_auto.sh
if [ $? = 99 ] ; then
   $SETVAL "ALSA" 1
   /etc/init.d/alsasound stop
else
   $SETVAL "ALSA" 0
fi   

ALL_MODS="$(sh /_config/bin/detect_modules.sh 2>&1)"
echo "Checking for ATIUSB"
if [ "$(echo "$ALL_MODS"| grep "^lirc_atiusb")" != "" ] ; then
   $SETVAL LIRC MedionX10
fi

echo "Checking for iMON"
if [ "$(echo "$ALL_MODS" | grep "iMON PAD")" != "" ] || [ "$(echo "$ALL_MODS" | grep "^imon_vfd")" != "" ] ; then
   autoload "imon_vfd"
   autoload "lirc_imon"
   sed -i /etc/LCDd.conf -e "s/^Driver=.*/Driver=imon/"
fi